/*
 rudi5mosaic

   Authors:  Hassan Siddiqui - XMM-SOC
         
   Description: Combines rudi-5 sub-images into a single mosaic
 
   Comments: 
 
  $Id: rudi5mosaic.cc,v 1.2 2000/06/22 14:51:06 hsiddiqu Exp $
*/

/* Sas-specific */
#include "Tracer.h"
#include "errstr.h"
#include "param.h"
#include "Dal.h"

/* Local */
#include "rudi5mosaic.h"

/* A hack. Need to include NumericalRange in eg utils */
#ifndef FLT32_MIN
#  define FLT32_MIN 1.17549435e-38F
#endif


void rudi5mosaic ()
{
  PROC;


  /******************************************************* 
   * Get bounds of mosaic window, in detector coordinates
   *******************************************************/

  unsigned xmin = intParameter("xmin");
  unsigned dx   = intParameter("dx");
  unsigned ymin = intParameter("ymin");
  unsigned dy   = intParameter("dy");

  unsigned xmax = xmin + dx - 1;
  unsigned ymax = ymin + dy - 1 ;

  longVec size;
  size.push_back(dy);
  size.push_back(dx);


  /********************************************************** 
   * Iterate through input sub-images, rejecting those
   *  sub images which do not overlap window, or are otherwise
   *  invalid from subsequent processing.
   ***********************************************************/

  vector<string> subimages_in = stringListParameter("imagearrays");
  vector<string> subimages;

  for (unsigned i = 0; i < subimages_in.size(); ++i)
    {
      errstr << "Accessing " << subimages_in[i] << " ..."
	     << Sas::message(Msg::App, Msg::Sparse);

      DataSet * set = dataSetServer->open(subimages_in[i], Dal::Read);
      
      Array * array = set->array(subimages_in[i]);
      
      /* Get size info */
      longVec arraysize = array->size();
      unsigned xsubsize = arraysize[1];
      unsigned ysubsize = arraysize[0]; // Note: Y comes first in arrays

      /* Check if x,y size matches array size */
      unsigned xwinsize = array->attribute("WINDOWDX")->asInt();
      unsigned ywinsize = array->attribute("WINDOWDY")->asInt();

      if (xsubsize != xwinsize || ysubsize != ywinsize)
	{
	  errstr << "Size for subimage "
		 << subimages_in[i] << " is ["
		 << xsubsize << "," << ysubsize
		 << "] which differs from that inferred from WINDOWDX, WINDOWDY ["
		 << xwinsize << "," << ywinsize
		 << "]. Will not include this image in mosaic."
		 << Sas::warning("subimageSizeMismatch");

	  /* Cleanup and move to next loop */
	  dataSetServer->close(set);
	  continue;
	}

      /* If subimage is not enclosed by or overlaps
	 the mosaic window at all, skip subimage entirely. */

      unsigned xsubmin = array->attribute("WINDOWX0")->asInt();
      unsigned ysubmin = array->attribute("WINDOWY0")->asInt();
      unsigned xsubmax = xsubmin+xsubsize-1;
      unsigned ysubmax = ysubmin+ysubsize-1;

      errstr << "Subimage " << subimages_in[i]
	     << " covers region ["
	     << xsubmin << "," << ysubmin << "] to ["
	     << xsubmax << "," << ysubmax
	     << "]"
	     << Sas::message(Msg::App, Msg::Noisy);

      if (!( 
	    ( xsubmin >= xmin && xsubmin <= xmax && ysubmin >= ymin && ysubmin <= ymax ) ||
	    ( xsubmin >= xmin && xsubmin <= xmax && ysubmax >= ymin && ysubmax <= ymax ) ||
	    ( xsubmax >= xmin && xsubmax <= xmax && ysubmin >= ymin && ysubmin <= ymax ) ||
	    ( xsubmax >= xmin && xsubmax <= xmax && ysubmax >= ymin && ysubmax <= ymax )
	    ))
	{
	  errstr << "Subimage " << subimages_in[i]
		 << "covers an area in det coords ["
		 << xsubmin << "," << ysubmin << "] to ["
		 << xsubmax << "," << ysubmax
		 << "] which is outside bounds of mosaic window (["
		 << xmin << "," << ymin << "] to ["
		 << xmax << "," << ymax
		 << "]). Will not include this image in mosaic."
		 << Sas::warning("subimageBoundInvalid");

	  /* Cleanup and move to next loop */
	  dataSetServer->close(set);
	  continue;
	}

      /* Cleanup and add subimage to stack of valid images */
      dataSetServer->close(set);
      subimages.push_back(subimages_in[i]);
    }

  /******************************************
   * Define small central window dimensions 
   ******************************************/

  unsigned small_xmin;
  unsigned small_dx;
  unsigned small_ymin;
  unsigned small_dy;

  if (booleanParameter("withsmallblock"))
    {
      string smallarrayname = stringParameter("smallblock");
      errstr << "Accessing " 
	     << smallarrayname
	     << " ..."
	     << Sas::message(Msg::App, Msg::Sparse);

      DataSet * set = 
	dataSetServer->open(smallarrayname, Dal::Read);
      
      Array * array = set->array(smallarrayname);
      
      small_xmin = array->attribute("WINDOWX0")->asInt();
      small_dy   = array->attribute("WINDOWY0")->asInt();
      small_ymin = array->attribute("WINDOWDX")->asInt();
      small_dy   = array->attribute("WINDOWDY")->asInt();

      dataSetServer->close(set);
      }
  else
    {
      small_xmin = intParameter("smallxmin");
      small_ymin = intParameter("smallymin");
      small_dx   = intParameter("smalldx");
      small_dy   = intParameter("smalldy");
    }
  unsigned small_xmax = small_xmin + small_dx - 1;
  unsigned small_ymax = small_ymin + small_dy - 1 ;


  /*****************************
   * Get exposure
   *****************************/
  real32 exposure = realParameter("exposure");


  /**********************************************************
   *
   * Create arrays for raw and exposure C++ matrices
   *
   ***********************************************************/

  errstr << "Creating and initialising raw matrix .."
	 << Sas::message(Msg::App, Msg::Sparse);
  vector< vector<int32> > rawmatrix;
  rawmatrix.resize(dx);
  for (unsigned x = 0; x < rawmatrix.size(); ++x)
    {
      rawmatrix[x].resize(dy);
      for (unsigned y = 0; y < rawmatrix[x].size(); ++y)
	{
	  rawmatrix[x][y] = 0;
	}
    }

  errstr << "Creating and initialising exp matrix .."
	 << Sas::message(Msg::App, Msg::Sparse);
  vector< vector<real32> > expmatrix;
  expmatrix.resize(dx);
  for (unsigned x = 0; x < expmatrix.size(); ++x)
    {
      expmatrix[x].resize(dy);
      for (unsigned y = 0; y < expmatrix[x].size(); ++y)
	{
	  expmatrix[x][y] = 0;
	}
    }

  /**********************************************************
   *
   * Iterate through input sub-images, 
   *  populating raw image and exposure maps accordingly.
   *
   ***********************************************************/

  for (unsigned i = 0; i < subimages.size(); ++i)
    {
      errstr << "Accessing " << subimages[i] << "..."
	     << Sas::message(Msg::App, Msg::Sparse);

      DataSet * set = dataSetServer->open(subimages[i], Dal::Read);
      
      Array * array = set->array(subimages[i]);

      longVec arraysize = array->size();
      unsigned ysubsize = arraysize[0];
      unsigned xsubsize = arraysize[1];
      
      /* Get lower bounds to image */
      unsigned xsubmin = array->attribute("WINDOWX0")->asInt();
      unsigned ysubmin = array->attribute("WINDOWY0")->asInt();

      /* Flag if this is the central low res window, as the region
	 corresponding to the small central window needs to be
	 removed. */
      
      bool central_low_res = false;
      
      if ( (int)xsubmin == intParameter("cenlowresxmin") && 
	   (int)ysubmin == intParameter("cenlowresymin") )
	{
	  errstr << 
	    "Current sub image is the central low res window. Will ignore region corresponding to small central window "
		 << Sas::message(Msg::App, Msg::Sparse);

	  central_low_res = true;
	}
	 
      /* Get array data */
      ArrayData * arraydata = array->data();
      int32 * data = arraydata->int32Data();

      /* Iterate though pixels */
      for (unsigned x = 0; x < xsubsize; ++x)
	{
	  for (unsigned y=0; y < ysubsize; ++y)
	    {
	      /* Compute detector coordinates */
	      unsigned xdet=xsubmin+x;
	      unsigned ydet=ysubmin+y;

	      if ( xdet < xmin || xdet > xmax || 
		   ydet < ymin || ydet > ymax )
		{
		  errstr 
		    << "Subimage pixel ["
		    << x << "," << y
		    << "] maps to a det coord ["
		    << xdet << "," << ydet
		    << "] which is out of bounds of mosaic ["
		    << xmin << "," << ymin << "] to ["
		    << xmax << "," << ymax 
		    << "]. Ignoring pixel "
		    << Sas::warning("subimagePixelOutOfBounds");

		  continue;
		}

	      if (central_low_res && 
		  small_xmin <= xdet && xdet <= small_xmax &&
		  small_ymin <= ydet && ydet <= small_ymax)
		{
		  continue;
		}

	      unsigned xpos = xdet - xmin;
	      unsigned ypos = ydet - ymin;
	      
	      /* Populate raw image */
	      int32 raw = rawmatrix[xpos][ypos];
	      int32 image = data[x + (y*xsubsize)];
	      
//  	      errstr 
//  		<< "Value of subimage[" 
//  		<< x << "," << y 
//  		<< "] is " << image << " ."
//  		<< " Replacing rawmatrix[" 
//  		<< xpos << "," << ypos
//  		<< "] value of " << raw
//  		<< " with " << raw + image
//  		<< Sas::message(Msg::App, Msg::Verbose);

	      rawmatrix[xpos][ypos] = raw + image;

	      /* Populate exposure map */
	      real32 exp = expmatrix[xpos][ypos];
	      
//  	      errstr
//  		<< " Replacing expmatrix[" 
//  		<< xpos << "," << ypos
//  		<< "] value of " << exp
//  		<< " with " << exp + exposure
//  		<< Sas::message(Msg::App, Msg::Verbose);

	      expmatrix[xpos][ypos] = exp + exposure;
	    }
	}
      
      /* Cleanup */
      delete arraydata;
      dataSetServer->close(set);
    }


  /**********************************************************
   *
   * Now create corrected matrix
   *
   ***********************************************************/


  errstr << "Creating corr matrix .."
	 << Sas::message(Msg::App, Msg::Sparse);
  vector< vector<real32> > corrmatrix;
  corrmatrix.resize(dx);
  for (unsigned x = 0; x < corrmatrix.size(); ++x)
    {
      corrmatrix[x].resize(dy);
      for (unsigned y = 0; y < corrmatrix[x].size(); ++y)
	{
	  int32 raw = rawmatrix[x][y];
	  real32 exp = expmatrix[x][y];

	  real32 corr;
	  
	  if (exp <= FLT32_MIN)
	    {
//    	      errstr << "Exposure map entry for position ["
//    		     << x << "," << y
//    		     << "] is zero. Setting corrected image value to zero "
//  		//  	      		     << Sas::warning("expMapZero");
//    		     << Sas::message(Msg::App, Msg::Verbose);
	      
	      corr = 0;
	      
	    }
	  else
	    {
	      corr = double(raw)/exp;
	    }
	  
	  corrmatrix[x][y] = corr;
	}
    }
  


  /*****************************
   *
   * Write data to DAL datasets
   *
   *****************************/

  /** Corrected image */
  string corrarrayname  = stringParameter("corrarray");
  errstr << "Opening " << corrarrayname << " .."
	 << Sas::message(Msg::App, Msg::Sparse);
  DataSet * corrset  = dataSetServer->open(corrarrayname, Dal::Create);
  Array * corrarray  = corrset->addArray(corrarrayname,
				       Dal::Real32,
				       size);
  ArrayData * corrarraydata = corrarray->data();
  real32 * corrdata = corrarraydata->real32Data();

  for (unsigned x = 0; x < dx; ++x)
    {
      for (unsigned y = 0; y < dy; ++y)
	{
	  corrdata[x+(y*dx)] = corrmatrix[x][y];
	}
    }

  delete corrarraydata;
  dataSetServer->close(corrset);

  /** Raw image */
  string rawarrayname  = stringParameter("rawarray");
  errstr << "Opening " << rawarrayname << " .."
	 << Sas::message(Msg::App, Msg::Sparse);
  DataSet * rawset  = dataSetServer->open(rawarrayname, Dal::Create);
  Array * rawarray  = rawset->addArray(rawarrayname,
				       Dal::Int32,
				       size);
  ArrayData * rawarraydata = rawarray->data();
  int32 * rawdata = rawarraydata->int32Data();

  for (unsigned x = 0; x < dx; ++x)
    {
      for (unsigned y = 0; y < dy; ++y)
	{
	  rawdata[x+(y*dx)] = rawmatrix[x][y];
	}
    }

  delete rawarraydata;
  dataSetServer->close(rawset);


  /** Exposure map */
  string exparrayname  = stringParameter("exparray");
  errstr << "Opening " << exparrayname << " .."
	 << Sas::message(Msg::App, Msg::Sparse);
  DataSet * expset  = dataSetServer->open(exparrayname, Dal::Create);
  Array * exparray  = expset->addArray(exparrayname,
				       Dal::Real32,
				       size);
  ArrayData * exparraydata = exparray->data();
  real32 * expdata = exparraydata->real32Data();

  for (unsigned x = 0; x < dx; ++x)
    {
      for (unsigned y = 0; y < dy; ++y)
	{
	  expdata[x+(y*dx)] = expmatrix[x][y];
	}
    }

  delete exparraydata;
  dataSetServer->close(expset);


  ENDPROC;
}
